""" Module for moPepGen parsers from different variant sources. """
from moPepGen.parser.REDItoolsParser import REDItoolsRecord
from moPepGen.parser import REDItoolsParser, VEPParser, STARFusionParser, \
	FusionCatcherParser, RMATSParser, CIRCexplorerParser, ArribaParser
